package edu.sc.seis.cormorant;

import java.util.Properties;
import org.omg.CORBA.ORB;
import org.omg.CORBA.Policy;
import org.omg.CORBA.ORBPackage.InvalidName;
import org.omg.PortableServer.IdAssignmentPolicyValue;
import org.omg.PortableServer.ImplicitActivationPolicyValue;
import org.omg.PortableServer.LifespanPolicyValue;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAHelper;
import org.omg.PortableServer.POAManager;
import org.omg.PortableServer.RequestProcessingPolicyValue;
import org.omg.PortableServer.Servant;
import org.omg.PortableServer.ServantRetentionPolicyValue;
import org.omg.PortableServer.POAPackage.AdapterAlreadyExists;
import org.omg.PortableServer.POAPackage.InvalidPolicy;
import edu.sc.seis.fissuresUtil.namingService.FissuresNamingService;

/**
 * @author crotwell Created on Dec 20, 2004
 */
public abstract class AbstractController implements Controller {

    public AbstractController(Properties confProps,
                              String serverPropName,
                              ORB orb) {
        this.serverPropName = serverPropName;
        this.configProps = confProps;
        this.serverName = configProps.getProperty(getPropertryPrefix() + "name");
        this.serverDNS = configProps.getProperty(getPropertryPrefix() + "dns");
        this.altServerName = configProps.getProperty(getPropertryPrefix()
                + "alternate.name");
        this.altServerDNS = configProps.getProperty(getPropertryPrefix()
                + "alternate.dns");
        if(serverName == null) {
            throw new NullPointerException("serverName cannot be null in properties as "
                    + getPropertryPrefix() + "name");
        }
        if(serverDNS == null) {
            throw new NullPointerException("serverDNS cannot be null in properties as "
                    + getPropertryPrefix() + "dns");
        }
    }

    public void start(ORB orb, FissuresNamingService ns) throws Exception {
        Servant servant = getServant();
        if(servant == null) {
            throw new NullPointerException("Servant cannot be NULL: "
                    + serverPropName);
        }
        org.omg.CORBA.Object obj = servant._this_object(orb);
        try {
        ns.rebind(getServerDNS(), getServerName(), obj, getInterfaceName());
        primaryBound = true;
        logger.info("Bound to NameService " + getServerDNS() + "/" + getInterfaceName() + "/"
                + getServerName() + " in " + ns.getNameServiceCorbaLoc());
        if(hasAlternateBinding()) {
            ns.rebind(getAlternateServerDNS(),
                      getAlternateServerName(),
                      obj,
                      getInterfaceName());
            alternateBound = true;
            logger.info("Bound to alternate NameService " + getAlternateServerDNS() + "/"
                    + getInterfaceName() + "/" + getAlternateServerName());
        }
        } catch (org.omg.CORBA.SystemException e) {
            logger.error("Unable to register with name service. "+configProps.get(Start.CORMORANT_NAME_SERVER), e);
            throw e;
        }
    }

    private boolean hasAlternateBinding() {
        return getAlternateServerDNS() != null
                && getAlternateServerName() != null
                && getAlternateServerDNS().length() != 0
                && getAlternateServerName().length() != 0;
    }

    public void stop(ORB orb, FissuresNamingService ns) throws Exception {
        if(orb != null) {
            // unregister with name service
            if(primaryBound) {
                ns.unbind(getServerDNS(), getInterfaceName(), getServerName());
                logger.info("Unbound " + getServerDNS() + "/"
                        + getInterfaceName() + "/" + getServerName() + " from "
                        + ns.getNameServiceCorbaLoc());
                primaryBound = false;
            }
            if(alternateBound) {
                ns.unbind(getAlternateServerDNS(),
                          getInterfaceName(),
                          getAlternateServerName());
                logger.info("Unbound " + getAlternateServerDNS() + "/"
                        + getInterfaceName() + "/" + getAlternateServerName());
                alternateBound = false;
            }
        }
    }

    public void destroy() throws Exception {}

    /**
     * Should be overridden if the interface is not one of the standard DHI
     * server types. Currently this includes EventDC, NetworkDC, PlottableDC and
     * DataCenter.
     */
    public String getInterfaceName() {
        return FissuresNamingService.getInterfaceName(getServant());
    }

    public abstract Servant getServant();

    public Properties getConfigProps() {
        return configProps;
    }

    public String getServerPropName() {
        return serverPropName;
    }

    public String getPropertryPrefix() {
        return Start.CORMORANT_SERVERS + "." + getServerPropName() + ".";
    }

    public String getServerDNS() {
        return serverDNS;
    }

    public String getServerName() {
        return serverName;
    }

    public String getAlternateServerDNS() {
        return altServerDNS;
    }

    public String getAlternateServerName() {
        return altServerName;
    }

    public POA createPOA(ORB orb) throws InvalidName, AdapterAlreadyExists,
            InvalidPolicy {
        POA rootPOA = POAHelper.narrow(orb.resolve_initial_references("RootPOA"));
        POAManager manager = rootPOA.the_POAManager();
        return rootPOA.create_POA(serverPropName,
                                  manager,
                                  createPolicy(rootPOA));
    }

    public static Policy[] createPolicy(POA poa) {
        return new Policy[] {poa.create_lifespan_policy(LifespanPolicyValue.PERSISTENT),
                             poa.create_id_assignment_policy(IdAssignmentPolicyValue.USER_ID),
                             poa.create_implicit_activation_policy(ImplicitActivationPolicyValue.NO_IMPLICIT_ACTIVATION),
                             poa.create_request_processing_policy(RequestProcessingPolicyValue.USE_SERVANT_MANAGER),
                             poa.create_servant_retention_policy(ServantRetentionPolicyValue.NON_RETAIN)};
    }

    private String serverDNS, serverName, altServerDNS, altServerName;

    private String serverPropName;

    private Properties configProps;

    private boolean primaryBound = false, alternateBound = false;

    private static final org.apache.log4j.Logger logger = org.apache.log4j.Logger.getLogger(AbstractController.class);
}