package edu.sc.seis.cormorant.plottable;

import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import org.apache.log4j.Logger;
import org.omg.CORBA.NO_IMPLEMENT;
import org.omg.CORBA.NO_MEMORY;
import org.omg.CORBA.UNKNOWN;
import edu.iris.Fissures.Dimension;
import edu.iris.Fissures.Plottable;
import edu.iris.Fissures.Time;
import edu.iris.Fissures.IfEvent.EventAccess;
import edu.iris.Fissures.IfNetwork.ChannelId;
import edu.iris.Fissures.IfNetwork.NetworkId;
import edu.iris.Fissures.IfPlottable.PlottableDCPOA;
import edu.iris.Fissures.IfPlottable.UnsupportedDimension;
import edu.iris.Fissures.IfSeismogramDC.DataCenterOperations;
import edu.iris.Fissures.IfSeismogramDC.LocalSeismogram;
import edu.iris.Fissures.IfSeismogramDC.RequestFilter;
import edu.iris.Fissures.model.MicroSecondDate;
import edu.iris.Fissures.model.TimeInterval;
import edu.iris.Fissures.model.UnitImpl;
import edu.iris.Fissures.network.ChannelIdUtil;
import edu.iris.Fissures.seismogramDC.LocalSeismogramImpl;
import edu.iris.Fissures.seismogramDC.RequestFilterUtil;
import edu.iris.dmc.seedcodec.CodecException;
import edu.sc.seis.fissuresUtil.cache.BulletproofVestFactory;
import edu.sc.seis.fissuresUtil.chooser.ClockUtil;
import edu.sc.seis.fissuresUtil.database.ConnMgr;
import edu.sc.seis.fissuresUtil.database.ConnectionCreator;
import edu.sc.seis.fissuresUtil.database.NotFound;
import edu.sc.seis.fissuresUtil.database.network.JDBCChannel;
import edu.sc.seis.fissuresUtil.database.network.JDBCNetwork;
import edu.sc.seis.fissuresUtil.database.plottable.JDBCPlottable;
import edu.sc.seis.fissuresUtil.database.plottable.PlottableChunk;
import edu.sc.seis.fissuresUtil.display.MicroSecondTimeRange;
import edu.sc.seis.fissuresUtil.display.SimplePlotUtil;
import edu.sc.seis.fissuresUtil.exceptionHandler.GlobalExceptionHandler;
import edu.sc.seis.fissuresUtil.simple.Initializer;
import edu.sc.seis.fissuresUtil.time.CoverageTool;

public class PlottableImpl extends PlottableDCPOA {

    public PlottableImpl(ConnectionCreator connCreator,
                         String dcDNS,
                         String dcName,
                         int dcTries) throws SQLException {
        this(getDataCenter(dcDNS, dcName, dcTries), connCreator);
    }

    public PlottableImpl(boolean cacheOnly,
                         String dcDNS,
                         String dcName,
                         int dcTries,
                         ConnectionCreator connCreator) throws SQLException {
        this((cacheOnly ? null : getDataCenter(dcDNS, dcName, dcTries)),
             connCreator);
    }

    public PlottableImpl(DataCenterOperations dc, ConnectionCreator connCreator)
            throws SQLException {
        this.dataCenter = dc;
        Connection conn = connCreator.createConnection();
        plottableDB = new JDBCPlottable(conn, connCreator.getType());
        netDB = new JDBCNetwork(conn);
        chanDB = new JDBCChannel(conn);
    }

    public boolean custom_sizes() {
        return false;
    }

    public Plottable[] get_plottable(RequestFilter request, Dimension pixel_size)
            throws UnsupportedDimension {
        long requestNumber = getRequestCount();
        logger.info(requestNumber + ": got request "
                + RequestFilterUtil.toString(request) + " at pixel size "
                + pixel_size.width);
        MicroSecondDate start = ClockUtil.now();
        int width = pixel_size.width;
        MicroSecondTimeRange requestRange = new MicroSecondTimeRange(request);
        int requestedPixelsPerDay = (int)(Math.floor(getDays(requestRange)) * width);
        boolean acceptableWidth = false;
        for(int i = 0; i < DAY_DIMS.length; i++) {
            if(DAY_DIMS[i].width == requestedPixelsPerDay) {
                acceptableWidth = true;
            }
        }
        if(!acceptableWidth) {
            throw new UnsupportedDimension(getCacheOnlyForLogger()
                    + " request #" + requestNumber
                    + ": Acceptable pixels per day are " + SUPPORTED_WIDTHS);
        }
        try {
            // start looking for better chanId than the one in the request
            // (IE we don't care about the start dates)
            // this may break if we hook this up to a seismogram server that
            // actually cares about dates.
            ChannelId chanId = getBestChannelIdFromDb(request);
            request.channel_id = chanId;
            RequestFilter[] filters = {request};
            // end of that long mess
            PlottableChunk[] chunks;
            synchronized(plottableDB) {
                chunks = plottableDB.get(requestRange,
                                         chanId,
                                         (int)(Math.floor(getDays(requestRange)) * width));
            }
            logger.debug(requestNumber + ": Got " + chunks.length
                    + " chunks for request #" + requestNumber + " "
                    + requestRange);
            MicroSecondTimeRange[] ranges = new MicroSecondTimeRange[chunks.length];
            for(int i = 0; i < chunks.length; i++) {
                ranges[i] = chunks[i].getTimeRange();
            }
            filters = CoverageTool.notCoveredIgnoreGaps(filters, ranges);
            Plottable[] newPlots = new Plottable[0];
            if(filters.length > 0 && dataCenter != null) {
                for(int i = 0; i < filters.length; i++) {
                    logger.debug(requestNumber + ": "
                            + RequestFilterUtil.toString(filters[i])
                            + " not covered by data in db");
                }
                // because some seismogram servers can't handle large requests,
                // we're breaking them up into requests <= six hours
                filters = breakUpFilters(filters,
                                         new TimeInterval(6.0, UnitImpl.HOUR));
                PlottableChunk[][] newChunks = new PlottableChunk[DAY_DIMS.length][0];
                for(int i = 0; i < filters.length; i++) {
                    try {
                        logger.debug(requestNumber + ": request # " + i
                                + " to data center: "
                                + RequestFilterUtil.toString(filters[i]));
                        LocalSeismogram[] seis = dataCenter.retrieve_seismograms(new RequestFilter[] {filters[i]});
                        for(int j = 0; j < seis.length; j++) {
                            LocalSeismogramImpl impl = (LocalSeismogramImpl)seis[j];
                            logger.debug(requestNumber
                                    + ": Got seismogram from "
                                    + new MicroSecondTimeRange(impl));
                        }
                        for(int j = 0; j < DAY_DIMS.length; j++) {
                            int curWidth = DAY_DIMS[j].width;
                            PlottableChunk[] seisChunks = makePlottables(seis,
                                                                         curWidth,
                                                                         requestRange);
                            PlottableChunk[] newNewChunks = new PlottableChunk[newChunks[j].length
                                    + seisChunks.length];
                            System.arraycopy(newChunks[j],
                                             0,
                                             newNewChunks,
                                             0,
                                             newChunks[j].length);
                            System.arraycopy(seisChunks,
                                             0,
                                             newNewChunks,
                                             newChunks[j].length,
                                             seisChunks.length);
                            newChunks[j] = newNewChunks;
                        }
                    } catch(NO_MEMORY e) {
                        logger.error(requestNumber + ": request # " + i
                                + " to data center: "
                                + RequestFilterUtil.toString(filters[i]), e);
                    }
                }
                newPlots = commitPlotsToDBAndFilterForWidth(newChunks,
                                                            requestedPixelsPerDay,
                                                            plottableDB);
            }
            Plottable[] all = new Plottable[chunks.length + newPlots.length];
            for(int i = 0; i < chunks.length; i++) {
                all[i] = chunks[i].getData();
            }
            System.arraycopy(newPlots, 0, all, chunks.length, newPlots.length);
            MicroSecondDate stop = ClockUtil.now();
            logger.debug(requestNumber + ": "
                    + stop.difference(start).convertTo(UnitImpl.SECOND)
                    + " to satisfy request for " + requestRange);
            return all;
        } catch(Throwable e) {
            GlobalExceptionHandler.handle("Exception occured while retrieving plottable using get_plottable. request "
                                                  + requestNumber
                                                  + ": "
                                                  + ChannelIdUtil.toString(request.channel_id)
                                                  + " from "
                                                  + new MicroSecondTimeRange(request.start_time,
                                                                             request.end_time),
                                          e);
            throw new UNKNOWN(e.toString());
        }
    }

    /**
     * look for better chanId than the one in the request (IE we don't care
     * about the start dates). this may break if we hook this up to a seismogram
     * server that actually cares about dates.
     */
    public ChannelId getBestChannelIdFromDb(ChannelId chanId) throws NotFound,
            SQLException {
        NetworkId[] nets;
        synchronized(netDB) {
            nets = netDB.getByCode(chanId.network_id.network_code);
        }
        NetworkId theNetId = null;
        if(nets.length > 0) {
            theNetId = nets[0];
        }
        if(theNetId != null) {
            ChannelId[] chanIds;
            synchronized(chanDB) {
                chanIds = chanDB.getIdsByCode(theNetId,
                                              chanId.station_code,
                                              chanId.site_code,
                                              chanId.channel_code);
            }
            ChannelId theChan = null;
            if(chanIds.length > 0) {
                theChan = chanIds[0];
            }
            if(theChan != null) {
                chanId = theChan;
            }
        }
        return chanId;
    }

    public ChannelId getBestChannelIdFromDb(RequestFilter request)
            throws NotFound, SQLException {
        return getBestChannelIdFromDb(request.channel_id);
    }

    protected static PlottableChunk[] makePlottables(LocalSeismogram[] seis,
                                                     int pixelsPerDay,
                                                     MicroSecondTimeRange time)
            throws CodecException, SQLException, IOException {
        List chunks = new ArrayList();
        for(int i = 0; i < seis.length; i++) {
            LocalSeismogramImpl curSeis = (LocalSeismogramImpl)seis[i];
            if(curSeis.getNumPoints() > 0
                    && SimplePlotUtil.canMakeAtLeastOnePixel(seis[i],
                                                             pixelsPerDay)) {
                Plottable plott = SimplePlotUtil.makePlottable(curSeis,
                                                               time,
                                                               pixelsPerDay);
                if(plott.x_coor.length > 0) {
                    PlottableChunk chunk = new PlottableChunk(plott,
                                                              SimplePlotUtil.getDayPixelRange(seis[i],
                                                                                              pixelsPerDay,
                                                                                              SimplePlotUtil.getBeginningOfDay(time.getBeginTime()))
                                                                      .getMin(),
                                                              time.getBeginTime(),
                                                              pixelsPerDay,
                                                              curSeis.channel_id);
                    chunks.add(chunk);
                }
            }
        }
        return (PlottableChunk[])chunks.toArray(new PlottableChunk[0]);
    }

    protected static Plottable[] commitPlotsToDBAndFilterForWidth(PlottableChunk[][] chunks,
                                                                  int pixelsPerDay,
                                                                  JDBCPlottable db)
            throws SQLException, IOException {
        List plottable = new ArrayList();
        for(int i = 0; i < DAY_DIMS.length; i++) {
            int curWidth = DAY_DIMS[i].width;
            if(curWidth == pixelsPerDay) {
                for(int j = 0; j < chunks[i].length; j++) {
                    Plottable plott = chunks[i][j].getData();
                    if(plott.x_coor.length > 0) {
                        plottable.add(plott);
                    }
                }
            }
            if(chunks[i].length > 0) {
                synchronized(db) {
                    db.put(chunks[i]);
                }
            }
        }
        return (Plottable[])plottable.toArray(new Plottable[0]);
    }

    public synchronized Plottable[] get_for_day(ChannelId channel_id,
                                                int year,
                                                int jday,
                                                Dimension pixel_size)
            throws UnsupportedDimension {
        try {
            Time[] times = getTimes(jday, year);
            return get_plottable(new RequestFilter(channel_id,
                                                   times[0],
                                                   times[1]), pixel_size);
        } catch(UnsupportedDimension e) {
            throw e;
        } catch(Throwable e) {
            GlobalExceptionHandler.handle(getCacheOnlyForLogger()
                                                  + "Exception occured while retrieving plottable using get_for_day",
                                          e);
            throw new org.omg.CORBA.UNKNOWN(e.toString());
        }
    }

    public Dimension[] get_whole_day_sizes() {
        return DAY_DIMS;
    }

    public void shutdownDB() throws SQLException {
        ConnMgr.createConnection().createStatement().execute("SHUTDOWN");
    }

    private static double getDays(MicroSecondTimeRange tr) {
        return tr.getInterval().convertTo(UnitImpl.DAY).getValue();
    }

    private static final Time[] getTimes(int jday, int year) {
        DecimalFormat numberFormat = new DecimalFormat("000");
        String yearString = new Integer(year).toString();
        Time startTime = new Time(yearString + numberFormat.format(jday)
                + "J00:00:00.000Z", -1);
        Time endTime = new Time(yearString + numberFormat.format(jday + 1)
                + "J00:00:00.000Z", -1);
        return new Time[] {startTime, endTime};
    }

    public Dimension[] get_event_sizes() {
        throw new NO_IMPLEMENT();
    }

    public Plottable[] get_for_event(EventAccess event,
                                     ChannelId channel_id,
                                     Dimension pixel_size) {
        throw new NO_IMPLEMENT();
    }

    private static DataCenterOperations getDataCenter(String dns,
                                                      String name,
                                                      int tries) {
        return BulletproofVestFactory.vestSeismogramDC(dns,
                                                       name,
                                                       Initializer.getNS(),
                                                       tries);
    }

    private String getCacheOnlyForLogger() {
        if(dataCenter == null) {
            return "CachePlottableImpl: ";
        } else {
            return "DCPlottableImpl: ";
        }
    }

    private long getRequestCount() {
        if(requestCount <= Long.MAX_VALUE) {
            return requestCount++;
        }
        return 0;
    }

    public static RequestFilter[] breakUpFilters(RequestFilter[] filters,
                                                 TimeInterval maxRequestLength) {
        List okayFilters = new ArrayList();
        for(int i = 0; i < filters.length; i++) {
            RequestFilter curFilter = filters[i];
            MicroSecondDate start = new MicroSecondDate(curFilter.start_time);
            MicroSecondDate end = new MicroSecondDate(curFilter.end_time);
            TimeInterval reqLen;
            do {
                reqLen = end.subtract(start);
                MicroSecondDate breakPoint = end;
                if(reqLen.greaterThan(maxRequestLength)) {
                    breakPoint = start.add(maxRequestLength);
                }
                RequestFilter req = new RequestFilter(curFilter.channel_id,
                                                      start.getFissuresTime(),
                                                      breakPoint.getFissuresTime());
                okayFilters.add(req);
                curFilter.start_time = breakPoint.getFissuresTime();
                start = breakPoint;
            } while(reqLen.greaterThan(maxRequestLength));
        }
        return (RequestFilter[])okayFilters.toArray(new RequestFilter[0]);
    }

    private DataCenterOperations dataCenter;

    private static final Dimension[] DAY_DIMS = new Dimension[] {new Dimension(5400,
                                                                               1),
                                                                 new Dimension(6000,
                                                                               1)};

    private static final String SUPPORTED_WIDTHS;
    static {
        StringBuffer sb = new StringBuffer();
        for(int i = 0; i < DAY_DIMS.length; i++) {
            sb.append(DAY_DIMS[i].width + " ");
        }
        SUPPORTED_WIDTHS = sb.toString();
    }

    private static final Logger logger = Logger.getLogger(PlottableImpl.class);

    private JDBCPlottable plottableDB;

    private JDBCNetwork netDB;

    private JDBCChannel chanDB;

    private long requestCount = 0;
}